{-# LANGUAGE GADTs, GeneralizedNewtypeDeriving, RankNTypes #-}
module TicTacToe.Game
  (PlayerMonadT, Player(..), runPlayer, doGame)
where

import TicTacToe.Grid

import Control.Monad.State
import Control.Monad.Operational


data Request a where
  GetGrid   :: Request Grid
  TurnDone  :: (Grid, Maybe GridResult) -> Request ()
  GetResult :: Request (Maybe GridResult)

newtype PlayerMonadT m a = PMT (ProgramT Request m a)
  deriving (Functor, Monad, MonadTrans)

instance (MonadIO m) => MonadIO (PlayerMonadT m) where
  liftIO = lift . liftIO

-- | mstack and mtop are monads. mstack is a monad stack which can be run, up to mtop
data Player mstack mtop = Player {
  -- | Gets the mark of the player
  plMark      :: Mark,
  -- | Called when the player must play
  plTurn      :: Grid -> mstack Pos,
  -- | Called when player tries to play at a forbidden position
  plForbidden :: Pos -> mstack (),
  -- | Called when game has ended.
  plGameOver  :: GridResult -> mstack (),
  -- | Used to reach the topmost monad
  plLift      :: forall a. mtop a -> mstack a,
  -- | Used to run the monad stack the player runs in
  plRun       :: forall a. mstack a -> mtop a
}

instance Eq (Player m m') where
  p1 == p2 = plMark p1 == plMark p2

instance Show (Player m m') where
  show = show . plMark


playerLogic :: (Monad m) => Player m (PlayerMonadT m') -> m ()
playerLogic pl = do
  let toProg = plLift pl . PMT . singleton
  grid <- toProg GetGrid
  pos <- plTurn pl grid
  case checkCell grid (plMark pl) pos of
    Nothing -> do        -- The cell was already filled in
      plForbidden pl pos -- We signal the error
      playerLogic pl     -- We start the turn again
    Just newGridAndResult -> do
                         -- The cell has been successfully marked, so we got a new grid
      toProg $ TurnDone newGridAndResult
                         -- At this point, the interpreter will switch to the other player
      mbResult <- toProg $ GetResult
                         -- This player is back, and wants to know what's new
      case mbResult of
        Nothing  -> playerLogic pl
        Just res -> plGameOver pl res
 
runPlayer :: (Monad m) => Player m (PlayerMonadT m') -> PlayerMonadT m' ()
runPlayer pl = plRun pl $ playerLogic pl

doGame :: (Monad m) => Grid -> [PlayerMonadT m ()] -> m Grid
doGame initGrid players =
  mapM unwrap players >>= flip evalStateT (initGrid, Nothing) . eval
  where 
    unwrap (PMT pl) = viewT pl

    eval :: (Monad m) => [PromptT Request m ()] ->
                         StateT (Grid, Maybe GridResult) m Grid

    eval [] = liftM fst get

    eval ((Return _) : pls) = eval pls

    eval ((GetGrid :>>= pl) : pls) = do
      (grid, _) <- get
      p <- lift . viewT $ pl grid
      eval $ p : pls

    eval ((TurnDone (newGrid, mbResult) :>>= pl) : pls) = do
      put (newGrid, mbResult)
      p <- lift . viewT $ pl ()
      eval $ pls ++ [p]

    eval ((GetResult :>>= pl) : pls) = do
      (_, mbResult) <- get
      p <- lift . viewT $ pl mbResult
      eval $ p : pls

